"""git utilities for dealing with AutPy submissions.

Currently, this module uses subprocess to deal with git,
slowly being migrated from code written in the "orga" repo.

In the future, we might want to consider using a Python git wrapper
(possibly pure-Python even), in order to get improved performace
as well as a nicer API:

https://github.com/gitpython-developers/GitPython
https://www.pygit2.org/
https://wayfair-incubator.github.io/pygitops/why-pygitops/
https://www.dulwich.io/
https://github.com/FriendCode/gittle
"""

from __future__ import annotations

import fnmatch
import logging
import pathlib
import shlex
import subprocess

from autpy import paths

logger = logging.getLogger(__name__)


class Error(Exception):
    """Base class for all errors in this module."""


class GitError(Error):
    """Raised when the git subproces failed."""

    def __init__(self, error: subprocess.CalledProcessError) -> None:
        super().__init__()
        self.stdout = error.stdout
        self.stderr = error.stderr
        self.cmd = error.cmd
        self.returncode = error.returncode
        logger.debug("git error:")
        logger.debug(str(self))

    def __str__(self) -> str:
        cmd = " ".join(shlex.quote(arg) for arg in self.cmd)
        msg = f"Git command exited with code {self.returncode}: {cmd}"
        if self.stdout is not None:
            msg += f"\n\nstdout:\n{self.stdout}\nstderr:\n{self.stderr}"

        return msg


class NotAGitRepo(Error):
    """Raised in __init__ if no .git was found."""

    def __init__(self, path: pathlib.Path) -> None:
        super().__init__(
            f"No .git directory found in {path}. "
            "Use check_git=False if you're certain that this is okay."
        )


class NothingToCommit(Error):
    """Raised by commit_all() when there are no changes to commit."""


class UnknownRevision(Error):
    """Raised by commit_message() when a revision does not exist."""


class UnknownRemoteRef(Error):
    """Raised by pull_origin() if the remote ref does not exist."""


class Repo:
    @classmethod
    def submissions(cls) -> Repo:
        """Get a repo object for the submissions dir."""
        return cls(paths.submissions_dir())

    @staticmethod
    def _run_git(
        *args: str | pathlib.Path,
        cwd: pathlib.Path,
        capture_output: bool = True,
        check: bool = True,
    ) -> subprocess.CompletedProcess[str]:
        """Internal wrapper for running git."""
        try:
            proc = subprocess.run(
                ["git"] + list(args),
                cwd=cwd,
                text=True,
                capture_output=capture_output,
                check=check,
            )
        except subprocess.CalledProcessError as exc:
            raise GitError(exc)

        argstr = " ".join(shlex.quote(str(arg)) for arg in args)
        logger.debug(f"git {argstr}")
        if proc.stdout:
            logger.debug(f"stdout:\n{proc.stdout}")
        if proc.stderr:
            logger.debug(f"stderr:\n{proc.stderr}")

        return proc

    @classmethod
    def init(
        cls,
        path: pathlib.Path,
        *,
        initial_branch: str | None = None,
    ) -> Repo:
        """Create a new repo via 'git init' and return it."""
        path.mkdir(exist_ok=True)
        cmd = ["init"]
        if initial_branch is not None:
            cmd.append(f"--initial-branch={initial_branch}")
        cls._run_git(*cmd, cwd=path)
        return Repo(path)

    def __init__(self, path: pathlib.Path, *, check_git: bool = True) -> None:
        if check_git and not (path / ".git").exists():
            raise NotAGitRepo(path)

        self.path = path

    def git(
        self,
        *args: str | pathlib.Path,
        capture_output: bool = True,
        check: bool = True,
    ) -> subprocess.CompletedProcess[str]:
        """Run a git subprocess and return the process object."""
        return self._run_git(
            *args,
            cwd=self.path,
            capture_output=capture_output,
            check=check,
        )

    def __repr__(self) -> str:
        return f"Repo({self.path!r})"

    def git_stdout(self, *args: str) -> str:
        """Run a git subprocess and get the stripped stdout."""
        return self.git(*args).stdout.strip()

    def current_branch(self) -> str:
        return self.git_stdout("branch", "--show-current")

    def ls_files(
        self,
        branch: str = "HEAD",
        pattern: str | None = None,
    ) -> set[pathlib.PurePath]:
        try:
            stdout = self.git_stdout("ls-tree", "-r", "--name-only", branch)
        except GitError as e:
            if e.stderr == f"fatal: Not a valid object name {branch}\n":
                return set()
            raise  # pragma: no cover

        paths = set()
        for line in stdout.splitlines():
            path = pathlib.PurePath(line)
            if pattern is None or fnmatch.fnmatchcase(path.name, pattern):
                paths.add(path)

        return paths

    def commit_message(self, ref: str) -> str:
        try:
            return self.git_stdout("log", "--format=%B", "-n", "1", ref)
        except GitError as e:
            if e.stderr.startswith(
                f"fatal: ambiguous argument '{ref}': unknown revision or path not in "
                "the working tree.\n"
            ):
                raise UnknownRevision(ref)
            raise  # pragma: no cover

    def commit_all(self, message: str) -> str:
        """Commit and push all files in a repository.

        Returns the short SHA of the made commit.
        Used in taas server to take submissions.
        """
        changes = self.git_stdout("status", "--porcelain=v1")
        if not changes:
            raise NothingToCommit()

        self.git("add", "--all")
        self.git("commit", "-m", message)
        return self.head_short_sha()

    def head_short_sha(self) -> str:
        """Get the current short SHA of HEAD."""
        return self.git_stdout("rev-parse", "--short", "HEAD")

    def push_origin(self) -> None:
        """Push all changes in the current branch to origin."""
        self.git("push", "--set-upstream", "origin", self.current_branch())

    def pull_origin(self) -> None:
        """Pull all changes in the current branch from origin."""
        ref = self.current_branch()
        try:
            self.git("pull", "--ff-only", "--set-upstream", "origin", ref)
        except GitError as e:
            if e.stderr.strip() == f"fatal: couldn't find remote ref {ref}":
                raise UnknownRemoteRef(ref)
            raise  # pragma: no cover

    def clone_to(
        self,
        dest_path: pathlib.Path,
        *,
        bare: bool = False,
    ) -> Repo:
        """Clone the current repo via 'git clone' and return the clone."""
        cmd = ["clone"]
        if bare:
            cmd.append("--bare")
        cmd += [str(self.path), str(dest_path)]
        self._run_git(*cmd, cwd=dest_path.parent)
        return Repo(dest_path, check_git=not bare)

    def config(self, key: str) -> str | None:
        """Get the given config value, or raise KeyError if it doesn't exist."""
        try:
            return self.git_stdout("config", key)
        except GitError as e:
            if e.returncode == 1:
                raise KeyError(key) from None
            raise  # pragma: no cover

    def config_set(self, config: dict[str, str]) -> None:
        """Run 'git config' commands for the given keys/values."""
        for key, value in config.items():
            self.git("config", key, value)
