"""Dealing with repository paths.

Paths can be overridden by the following environment variables:

- AUTPY_ROOT_PATH: Root directory containing all the other repos
- AUTPY_SUBMISSIONS_PATH: Directory containing the submissions repo
- AUTPY_NOTEBOOKS_PATH: Directory containing the (deployed) notebooks repo
"""

import functools
import os
import pathlib
from typing import Callable

from autpy import utils


class _path_func:

    """Decorator to add env-var overriding and checks to functions below."""

    def __init__(self, name: str, *, is_git: bool = True) -> None:
        self.name = name
        self.is_git = is_git

    def __call__(self, func: Callable[[], pathlib.Path]) -> Callable[[], pathlib.Path]:
        @functools.wraps(func)
        def wrapped() -> pathlib.Path:
            envvar = f"AUTPY_{self.name.upper()}_PATH"
            override = os.environ.get(envvar, None)

            path = pathlib.Path(override) if override is not None else func()
            source = "auto-detected" if override is None else f"${envvar}"

            if not path.exists():
                raise ValueError(
                    f"AutPy {self.name} directory not found at {path} "
                    f"(source: {source})"
                )
            if self.is_git and not (path / ".git").exists():
                raise ValueError(
                    f"Expected AutPy {self.name} directory to be a git repository, but "
                    f"no .git folder found in {path} (source: {source})"
                )

            return path.resolve()

        return wrapped


@_path_func("root", is_git=False)
def root_dir() -> pathlib.Path:
    """Find the the autpy root directory."""
    repo_path = pathlib.Path(__file__).parents[2]

    # sanity check
    if repo_path.name != "python-autpy" or not (repo_path / ".git").exists():
        # FIXME allow different names?
        raise ValueError(
            f"Expected python-autpy repo at {repo_path}, but found a different "
            "directory, or no git repository"
        )

    return repo_path.parent


@_path_func("submissions")
def submissions_dir() -> pathlib.Path:
    """Find the autpy submissions directory."""
    return root_dir() / f"submissions-{utils.semester_suffix()}"


@_path_func("notebooks")
def notebooks_dir() -> pathlib.Path:
    """Find the autpy notebooks directory."""
    return root_dir() / "notebooks"


@_path_func("development")
def development_dir() -> pathlib.Path:
    """Find the autpy development directory."""
    return root_dir() / "development"
