import logging
import pathlib
import re
import subprocess
import sys
import textwrap

import pytest

from autpy import git


@pytest.fixture
def clone_repo(
    repo: git.Repo, tmp_path: pathlib.Path, git_config: dict[str, str]
) -> git.Repo:
    clone_path = tmp_path / "repo-clone"
    clone = repo.clone_to(clone_path)
    clone.config_set(git_config)
    return clone


@pytest.fixture
def bare_clone_repo(
    repo: git.Repo, tmp_path: pathlib.Path, git_config: dict[str, str]
) -> git.Repo:
    clone_path = tmp_path / "repo-clone-bare"
    clone = repo.clone_to(clone_path, bare=True)
    clone.config_set(git_config)
    return clone


@pytest.fixture
def dummy_repo(tmp_path: pathlib.Path) -> git.Repo:
    """A git repo which doesn't actually exist."""
    return git.Repo(tmp_path, check_git=False)


@pytest.mark.parametrize("capture_output", [True, False])
def test_git_error_exception(capture_output: bool) -> None:
    src = " ;".join(
        [
            "import sys",
            'print("standard output")',
            'print("standard error", file=sys.stderr)',
            "sys.exit(2)",
        ]
    )
    with pytest.raises(subprocess.CalledProcessError) as excinfo:
        subprocess.run(
            [sys.executable, "-c", src],
            check=True,
            capture_output=capture_output,
            text=True,
        )

    exc = git.GitError(excinfo.value)

    expected = f"Git command exited with code 2: {sys.executable} -c '{src}'"
    if capture_output:
        expected += textwrap.dedent(
            """

            stdout:
            standard output

            stderr:
            standard error
            """
        )
    assert str(exc) == expected


def test_git(tmp_path: pathlib.Path) -> None:
    git_dir = tmp_path / ".git"
    assert not git_dir.exists()

    git.Repo(tmp_path, check_git=False).git("init")
    assert git_dir.exists()


def test_no_git_dir(tmp_path: pathlib.Path) -> None:
    message = (
        f"No .git directory found in {tmp_path}. "
        "Use check_git=False if you're certain that this is okay."
    )
    with pytest.raises(git.NotAGitRepo, match=re.escape(message)):
        git.Repo(tmp_path)


@pytest.mark.parametrize("capture", [True, False])
def test_capturing(dummy_repo: git.Repo, capture: bool) -> None:
    proc = dummy_repo.git("--version", capture_output=capture)
    if capture:
        assert proc.stdout.startswith("git version")
    else:
        assert proc.stdout is None


def test_git_stdout(dummy_repo: git.Repo) -> None:
    assert dummy_repo.git_stdout("--version").startswith("git version")


def test_git_logging(dummy_repo: git.Repo, caplog: pytest.LogCaptureFixture) -> None:
    with caplog.at_level(logging.DEBUG):
        dummy_repo.git("--version")

    assert caplog.messages[0] == "git --version"
    assert caplog.messages[1].startswith("stdout:\ngit version")


@pytest.mark.parametrize("capture_output", [True, False])
def test_error_command(dummy_repo: git.Repo, capture_output: bool) -> None:
    with pytest.raises(git.GitError) as excinfo:
        dummy_repo.git("gud", capture_output=capture_output)

    expected = "Git command exited with code 1: git gud"
    if capture_output:
        expected += textwrap.dedent(
            """

            stdout:

            stderr:
            git: 'gud' is not a git command. See 'git --help'.
            """
        )

    assert str(excinfo.value).startswith(expected)


def test_check_false(dummy_repo: git.Repo) -> None:
    proc = dummy_repo.git("gud", check=False)
    assert proc.returncode == 1


@pytest.mark.parametrize("checkout", [True, False])
def test_current_branch(repo: git.Repo, checkout: bool) -> None:
    if checkout:
        expected = "testbranch"
        repo.git("checkout", "-b", expected)
    else:
        expected = "master"

    assert repo.current_branch() == expected


class TestCommitAll:
    def test_nothing_to_commit(self, repo: git.Repo) -> None:
        with pytest.raises(git.NothingToCommit):
            repo.commit_all("test commit")

    def test_new_file(self, repo: git.Repo) -> None:
        (repo.path / "depp").touch()

        message = "test commit"
        sha = repo.commit_all(message)

        assert repo.commit_message(sha) == message

    def test_changed_file(self, repo: git.Repo) -> None:
        depp_file = repo.path / "depp"
        depp_file.touch()

        sha = repo.commit_all("empty file")

        depp_file.write_text("hello world", encoding="utf-8")

        message = "add lines"
        sha = repo.commit_all(message)

        assert repo.commit_message(sha) == message


def test_commit_message_missing_commit(repo: git.Repo) -> None:
    with pytest.raises(git.UnknownRevision, match="deadbeef"):
        repo.commit_message("deadbeef")


def test_push_origin(repo: git.Repo, bare_clone_repo: git.Repo) -> None:
    (repo.path / "depp").touch()
    message = "test commit"
    sha = repo.commit_all(message)

    # sanity check: commit should exist in original, not in clone yet
    assert repo.commit_message(sha) == message
    with pytest.raises(git.UnknownRevision):
        bare_clone_repo.commit_message(sha)

    repo.git("remote", "add", "origin", bare_clone_repo.path)
    repo.push_origin()
    assert bare_clone_repo.commit_message(sha) == message


def test_pull_origin(repo: git.Repo, clone_repo: git.Repo) -> None:
    (clone_repo.path / "depp").touch()
    message = "test commit"
    sha = clone_repo.commit_all(message)

    # sanity check: commit should exist in clone, not in original yet
    assert clone_repo.commit_message(sha) == message
    with pytest.raises(git.UnknownRevision):
        repo.commit_message(sha)

    repo.git("remote", "add", "origin", clone_repo.path)
    repo.pull_origin()
    assert repo.commit_message(sha) == message


def test_pull_origin_no_remote_branch(repo: git.Repo, clone_repo: git.Repo) -> None:
    repo.git("remote", "add", "origin", clone_repo.path)
    with pytest.raises(git.UnknownRemoteRef, match="master"):
        repo.pull_origin()


def test_submissions_repo(repo: git.Repo, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("AUTPY_SUBMISSIONS_PATH", str(repo.path))
    submissions_repo = git.Repo.submissions()
    assert submissions_repo.path == repo.path


def test_init(tmp_path: pathlib.Path) -> None:
    git_dir = tmp_path / ".git"
    assert not git_dir.exists()

    git.Repo.init(tmp_path)
    assert git_dir.exists()


def test_init_branch(tmp_path: pathlib.Path) -> None:
    branch = "zweig"
    r = git.Repo.init(tmp_path, initial_branch=branch)
    assert r.current_branch() == branch


@pytest.mark.parametrize("bare", [True, False])
def test_clone(repo: git.Repo, tmp_path: pathlib.Path, bare: bool) -> None:
    clone_path = tmp_path / "clone"

    clone = repo.clone_to(clone_path, bare=bare)

    marker = "HEAD" if bare else ".git"
    assert (clone.path / marker).exists()

    expected_remotes = [
        ("origin", str(repo.path), "(fetch)"),
        ("origin", str(repo.path), "(push)"),
    ]

    remotes = [
        tuple(line.split()) for line in clone.git_stdout("remote", "-v").splitlines()
    ]
    assert remotes == expected_remotes


def test_config(repo: git.Repo) -> None:
    key = "alias.autpy"
    value = "eggs"
    repo.config_set({key: value})
    assert repo.config(key) == value


def test_config_get_inexistent(repo: git.Repo) -> None:
    with pytest.raises(KeyError):
        repo.config("derpiness")


def test_repr(repo: git.Repo) -> None:
    assert repr(repo) == f"Repo({repo.path!r})"


class TestLsFiles:
    @pytest.fixture(autouse=True)
    def example_files(self, repo: git.Repo) -> set[pathlib.PurePath]:
        files = {pathlib.PurePath("eggs.txt"), pathlib.PurePath("bacon.jpg")}
        for path in files:
            (repo.path / path).touch()
        repo.commit_all("add files")
        return files

    def test_simple(self, example_files: set[pathlib.Path], repo: git.Repo) -> None:
        assert repo.ls_files() == example_files

    def test_other_branch(
        self, example_files: set[pathlib.Path], repo: git.Repo
    ) -> None:
        branch = "new-branch"
        repo.git("checkout", "-b", branch)
        (repo.path / "more").touch()
        repo.commit_all("add more")

        assert repo.ls_files() == example_files | {pathlib.PurePath("more")}
        assert repo.ls_files("master") == example_files

    def test_unknown_branch(self, repo: git.Repo) -> None:
        assert repo.ls_files("does-not-exist") == set()

    def test_pattern(self, repo: git.Repo) -> None:
        assert repo.ls_files(pattern="*.jpg") == {pathlib.PurePath("bacon.jpg")}
