import pathlib
import re
from typing import Callable

import pytest

from autpy import paths, utils


@pytest.mark.parametrize(
    "func, envvar, needs_git_dir",
    [
        (paths.root_dir, "AUTPY_ROOT_PATH", False),
        (paths.submissions_dir, "AUTPY_SUBMISSIONS_PATH", True),
        (paths.development_dir, "AUTPY_DEVELOPMENT_PATH", True),
        (paths.notebooks_dir, "AUTPY_NOTEBOOKS_PATH", True),
    ],
)
def test_overrides(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
    func: Callable[[], pathlib.Path],
    envvar: str,
    needs_git_dir: bool,
) -> None:
    monkeypatch.setenv(envvar, str(tmp_path))
    if needs_git_dir:
        (tmp_path / ".git").mkdir()
    assert func() == tmp_path


@pytest.mark.parametrize(
    "func, name, envvar",
    [
        (paths.submissions_dir, "submissions", "AUTPY_SUBMISSIONS_PATH"),
        (paths.notebooks_dir, "notebooks", "AUTPY_NOTEBOOKS_PATH"),
        (paths.development_dir, "development", "AUTPY_DEVELOPMENT_PATH"),
    ],
)
def test_overrides_no_git_dir(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
    func: Callable[[], pathlib.Path],
    name: str,
    envvar: str,
) -> None:
    monkeypatch.setenv(envvar, str(tmp_path))
    message = (
        f"Expected AutPy {name} directory to be a git repository, but no .git folder "
        f"found in {tmp_path} (source: ${envvar})"
    )
    with pytest.raises(ValueError, match=re.escape(message)):
        func()


@pytest.mark.parametrize(
    "func, name, envvar",
    [
        (paths.submissions_dir, "submissions", "AUTPY_SUBMISSIONS_PATH"),
        (paths.notebooks_dir, "notebooks", "AUTPY_NOTEBOOKS_PATH"),
        (paths.development_dir, "development", "AUTPY_DEVELOPMENT_PATH"),
    ],
)
def test_overrides_missing(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
    func: Callable[[], pathlib.Path],
    name: str,
    envvar: str,
) -> None:
    path = tmp_path / "does-not-exist"
    monkeypatch.setenv(envvar, str(path))
    message = f"AutPy {name} directory not found at {path} (source: ${envvar})"
    with pytest.raises(ValueError, match=re.escape(message)):
        func()


@pytest.mark.parametrize(
    "func, name",
    [
        (paths.root_dir, ""),
        (paths.submissions_dir, f"submissions-{utils.semester_suffix()}"),
        (paths.notebooks_dir, "notebooks"),
        (paths.development_dir, "development"),
    ],
)
def test_auto_detect(
    func: Callable[[], pathlib.Path],
    name: str,
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    python_autpy_path = tmp_path / "python-autpy"
    python_autpy_path.mkdir()
    (python_autpy_path / ".git").mkdir()

    src_path = python_autpy_path / "src" / "autpy"
    src_path.mkdir(parents=True)

    if name:
        dest_path = tmp_path / name
        dest_path.mkdir()
        (dest_path / ".git").mkdir()

    monkeypatch.setattr(paths, "__file__", str(src_path / "paths.py"))
    path = func()
    assert path.relative_to(tmp_path) == pathlib.Path(name)


@pytest.mark.parametrize(
    "folder_name, is_git",
    [
        ("python-autpy", False),
        ("not-python-autpy", True),
    ],
)
def test_root_auto_detect_invalid(
    folder_name: str,
    is_git: bool,
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    python_autpy_path = tmp_path / folder_name
    python_autpy_path.mkdir()
    if is_git:
        (python_autpy_path / ".git").mkdir()

    src_path = python_autpy_path / "src" / "autpy"
    src_path.mkdir(parents=True)

    monkeypatch.setattr(paths, "__file__", str(src_path / "paths.py"))
    message = (
        f"Expected python-autpy repo at {python_autpy_path}, but found a different "
        "directory, or no git repository"
    )
    with pytest.raises(ValueError, match=re.escape(message)):
        paths.root_dir()
