import pathlib

import pytest

from autpy import project, git


class TestProjectFiles:
    @pytest.fixture(autouse=True)
    def patch_repo(self, repo: git.Repo, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("AUTPY_SUBMISSIONS_PATH", str(repo.path))

    @pytest.fixture
    def project_files(self, repo: git.Repo) -> set[pathlib.PurePath]:
        project_path = repo.path / "project"
        project_path.mkdir()

        txt_path = project_path / "test.txt"
        txt_path.touch()

        grading_path = project_path / "grading.md"
        grading_path.touch()

        not_project_path = repo.path / "lab.ipynb"
        not_project_path.touch()

        repo.commit_all("Add project")

        return {
            txt_path.relative_to(repo.path),
            grading_path.relative_to(repo.path),
        }

    def test_has_project_empty_repo(self) -> None:
        assert not project.has_project("master")

    def test_project_files_empty_repo(self) -> None:
        assert project.get_project_files("master") == set()

    def test_has_project_with_files(self, project_files: set[pathlib.PurePath]) -> None:
        assert project.has_project("master")

    def test_project_files_with_files(
        self, project_files: set[pathlib.PurePath]
    ) -> None:
        assert project.get_project_files("master") == project_files

    def test_project_files_exclude_grading(
        self, project_files: set[pathlib.PurePath]
    ) -> None:
        expected = project_files - {pathlib.PurePath("project", "grading.md")}
        assert project.get_project_files("master", exclude_grading=True) == expected


@pytest.mark.parametrize(
    "path, expected",
    [
        (pathlib.PurePath("test.md"), False),
        (pathlib.PurePath("grading.jpg"), False),
        (pathlib.PurePath("grading.md"), True),
        (pathlib.PurePath("grading-BETA.md"), True),
    ],
)
def test_is_grading_path(path: pathlib.PurePath, expected: bool) -> None:
    assert project._is_grading_path(path) == expected


@pytest.mark.parametrize(
    "paths, expected",
    [
        ([pathlib.PurePath("test.md")], None),
        ([pathlib.PurePath("grading.md")], pathlib.PurePath("grading.md")),
        ([pathlib.PurePath("grading-BETA.md")], pathlib.PurePath("grading-BETA.md")),
        (
            [pathlib.PurePath("test.md"), pathlib.PurePath("grading.md")],
            pathlib.PurePath("grading.md"),
        ),
        (
            [pathlib.PurePath("test.md"), pathlib.PurePath("test2.md")],
            None,
        ),
    ],
)
def test_find_grading_path(
    paths: list[pathlib.PurePath], expected: pathlib.Path | None
) -> None:
    assert project.find_grading_path(paths) == expected
